/*
 * (C) Copyright 2000
 * Wolfgang Denk, DENX Software Engineering, wd@denx.de.
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

/*
 * Memory Functions
 *
 * Copied from FADS ROM, Dan Malek (dmalek@jlc.net)
 */

#include <common.h>
#include <command.h>
#if (CONFIG_COMMANDS & CFG_CMD_MMC)
#include <mmc.h>
#endif
#ifdef CONFIG_HAS_DATAFLASH
#include <dataflash.h>
#endif

#include "octeon_boot.h"

#if (CONFIG_COMMANDS & (CFG_CMD_MEMORY	| \
			CFG_CMD_I2C	| \
			CFG_CMD_ITEST	| \
			CFG_CMD_PCI	| \
			CMD_CMD_PORTIO	) )
int cmd_get_data_size(char* arg, int default_size)
{
	/* Check for a size specification .b, .w, .l, or .d.
	 */
	int len = strlen(arg);
	if (len > 2 && arg[len-2] == '.') {
		switch(arg[len-1]) {
		case 'b':
			return 1;
		case 'w':
			return 2;
		case 'd':
			return 4;
		case 'l':
			return 8;
		case 's':
			return -2;
		default:
			return -1;
		}
	}
	return default_size;
}
#endif

#if (CONFIG_COMMANDS & CFG_CMD_MEMORY)

#ifdef	CMD_MEM_DEBUG
#define	PRINTF(fmt,args...)	printf (fmt ,##args)
#else
#define PRINTF(fmt,args...)
#endif

static int mod_mem(cmd_tbl_t *, int, int, int, char *[]);

/* Display values from last command.
 * Memory modify remembered values are different from display memory.
 */
uint	dp_last_addr, dp_last_size;
uint	dp_last_length = 0x40;
uint	mm_last_addr, mm_last_size;

static	ulong	base_address = 0;
static	uint64_t	base_address64 = (1ull << 63);

/* Memory Display
 *
 * Syntax:
 *	md{.b, .w, .l} {addr} {len}
 */
#define DISP_LINE_LEN	16
int do_mem_md ( cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{
	ulong	addr, length;
	ulong	i, nbytes, linebytes;
	u_char	*cp;
	int	size;
	int rc = 0;

	/* We use the last specified parameters, unless new ones are
	 * entered.
	 */
	addr = dp_last_addr;
	size = dp_last_size;
	length = dp_last_length;

	if (argc < 2) {
		printf ("Usage:\n%s\n", cmdtp->usage);
		return 1;
	}

	if ((flag & CMD_FLAG_REPEAT) == 0) {
		/* New command specified.  Check for a size specification.
		 * Defaults to long if no or incorrect specification.
		 */
		if ((size = cmd_get_data_size(argv[0], 4)) < 0)
			return 1;

		/* Address is specified since argc > 1
		*/
		addr = simple_strtoul(argv[1], NULL, 16);
		addr += base_address;

		/* If another parameter, it is the length to display.
		 * Length is the number of objects, not number of bytes.
		 */
		if (argc > 2)
			length = simple_strtoul(argv[2], NULL, 16);
	}

	/* Print the lines.
	 *
	 * We buffer all read data, so we can make sure data is read only
	 * once, and all accesses are with the specified bus width.
	 */
	nbytes = length * size;
	do {
		char	linebuf[DISP_LINE_LEN];
		uint	*uip = (uint   *)linebuf;
		ushort	*usp = (ushort *)linebuf;
		u_char	*ucp = (u_char *)linebuf;
#ifdef CONFIG_HAS_DATAFLASH
		int rc;
#endif
		printf("%08lx:", addr);
		linebytes = (nbytes>DISP_LINE_LEN)?DISP_LINE_LEN:nbytes;

#ifdef CONFIG_HAS_DATAFLASH
		if ((rc = read_dataflash(addr, (linebytes/size)*size, linebuf)) == DATAFLASH_OK){
			/* if outside dataflash */
			/*if (rc != 1) {
				dataflash_perror (rc);
				return (1);
			}*/
			for (i=0; i<linebytes; i+= size) {
				if (size == 4) {
					printf(" %08x", *uip++);
				} else if (size == 2) {
					printf(" %04x", *usp++);
				} else {
					printf(" %02x", *ucp++);
				}
				addr += size;
			}

		} else {	/* addr does not correspond to DataFlash */
#endif
		for (i=0; i<linebytes; i+= size) {
			if (size == 4) {
				printf(" %08x", (*uip++ = *((uint *)addr)));
			} else if (size == 2) {
				printf(" %04x", (*usp++ = *((ushort *)addr)));
			} else {
				printf(" %02x", (*ucp++ = *((u_char *)addr)));
			}
			addr += size;
		}
#ifdef CONFIG_HAS_DATAFLASH
		}
#endif
		puts ("    ");
		cp = (uchar *)linebuf;
		for (i=0; i<linebytes; i++) {
			if ((*cp < 0x20) || (*cp > 0x7e))
				putc ('.');
			else
				printf("%c", *cp);
			cp++;
		}
		putc ('\n');
		nbytes -= linebytes;
		if (ctrlc()) {
			rc = 1;
			break;
		}
	} while (nbytes > 0);

	dp_last_addr = addr;
	dp_last_length = length;
	dp_last_size = size;
	return (rc);
}

int do_mem_mm ( cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{
	return mod_mem (cmdtp, 1, flag, argc, argv);
}
int do_mem_nm ( cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{
	return mod_mem (cmdtp, 0, flag, argc, argv);
}


int do_read_cmp ( cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{

	if ((argc != 3)) {
		printf ("Usage:\n%s\n", cmdtp->usage);
		return 1;
	}
    uint32_t addr;
    uint32_t val;
    uint32_t *ptr;
    addr = simple_strtoul(argv[1], NULL, 16);
    addr += base_address;

    val = simple_strtoul(argv[2], NULL, 16);
    printf("Looking for value other than 0x%x, addr: 0x%x\n", val, addr);
    ptr = (void *)addr;
    while (*ptr++ == val)
        ;
    ptr--;
    printf("Found 0x%x at addr %p\n", *ptr, ptr);

    return(0);
}

int do_read64 ( cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{

	if ((argc != 2)) {
		printf ("Usage:\n%s\n", cmdtp->usage);
		return 1;
	}
    uint64_t addr, val;
    addr = simple_strtoull(argv[1], NULL, 16);
    addr += base_address64;

    printf("attempting to read from addr: 0x%qx\n", addr);
    val = octeon_read64(addr);
    printf("0x%qx: 0x%qx\n", addr, val);

    return(0);
}
int do_read64b ( cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{

	if ((argc != 2)) {
		printf ("Usage:\n%s\n", cmdtp->usage);
		return 1;
	}
    uint64_t addr;
    uint8_t val;
    addr = simple_strtoull(argv[1], NULL, 16);
    addr += base_address64;

    printf("attempting to read from addr: 0x%qx\n", addr);
    val = octeon_read64_byte(addr);
    printf("0x%qx: 0x%x\n", addr, val);

    return(0);
}
int do_read64l ( cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{

	if ((argc != 2)) {
		printf ("Usage:\n%s\n", cmdtp->usage);
		return 1;
	}
    uint64_t addr;
    uint32_t val;
    addr = simple_strtoull(argv[1], NULL, 16);
    addr += base_address64;

    printf("attempting to read from addr: 0x%qx\n", addr);
    val = cvmx_read64_uint32(addr);
    printf("0x%qx: 0x%x\n", addr, val);

    return(0);
}

int do_write64 ( cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{

	if ((argc != 3)) {
		printf ("Usage:\n%s\n", cmdtp->usage);
		return 1;
	}
    uint64_t addr, val;
    addr = simple_strtoull(argv[1], NULL, 16);
    addr += base_address64;
    val = simple_strtoull(argv[2], NULL, 16);

    printf("writing 0x%qx to addr: 0x%qx\n", val, addr);

    octeon_write64(addr, val);
    return(0);
}

int do_write64b ( cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{

	if ((argc != 3)) {
		printf ("Usage:\n%s\n", cmdtp->usage);
		return 1;
	}
    uint64_t addr;
    uint8_t val;
    addr = simple_strtoull(argv[1], NULL, 16);
    addr += base_address64;
    val = simple_strtoull(argv[2], NULL, 16);

    printf("writing 0x%x to addr: 0x%qx\n", val, addr);

    octeon_write64_byte(addr, val);
    return(0);
}
int do_write64l ( cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{

	if ((argc != 3)) {
		printf ("Usage:\n%s\n", cmdtp->usage);
		return 1;
	}
    uint64_t addr;
    uint32_t val;
    addr = simple_strtoull(argv[1], NULL, 16);
    addr += base_address64;
    val = simple_strtoull(argv[2], NULL, 16);

    printf("writing 0x%x to addr: 0x%qx\n", val, addr);

    cvmx_write64_uint32(addr, val);
    return(0);
}


int do_mem_mw ( cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{
	ulong	addr, writeval, count;
	int	size;

	if ((argc < 3) || (argc > 4)) {
		printf ("Usage:\n%s\n", cmdtp->usage);
		return 1;
	}

	/* Check for size specification.
	*/
	if ((size = cmd_get_data_size(argv[0], 4)) < 1)
		return 1;

    if (size == 8)
    {
        uint64_t addr, writeval, count;
        addr = simple_strtoull(argv[1], NULL, 16);
        writeval = simple_strtoull(argv[2], NULL, 16);
        if (argc == 4) {
            count = simple_strtoull(argv[3], NULL, 16);
        } else {
            count = 1;
        }

        while (count-- > 0) {
            octeon_write64(addr, writeval);
            addr += 8;
        }
    }
	/* Address is specified since argc > 1
	*/
	addr = simple_strtoul(argv[1], NULL, 16);
	addr += base_address;

	/* Get the value to write.
	*/
	writeval = simple_strtoul(argv[2], NULL, 16);


	/* Count ? */
	if (argc == 4) {
		count = simple_strtoul(argv[3], NULL, 16);
	} else {
		count = 1;
	}

	while (count-- > 0) {
		if (size == 4)
			*((ulong  *)addr) = (ulong )writeval;
		else if (size == 2)
			*((ushort *)addr) = (ushort)writeval;
		else
			*((u_char *)addr) = (u_char)writeval;
		addr += size;
	}
	return 0;
}

int do_mem_cmp (cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{
	ulong	addr1, addr2, count, ngood;
	int	size;
	int     rcode = 0;

	if (argc != 4) {
		printf ("Usage:\n%s\n", cmdtp->usage);
		return 1;
	}

	/* Check for size specification.
	*/
	if ((size = cmd_get_data_size(argv[0], 4)) < 0)
		return 1;

	addr1 = simple_strtoul(argv[1], NULL, 16);
	addr1 += base_address;

	addr2 = simple_strtoul(argv[2], NULL, 16);
	addr2 += base_address;

	count = simple_strtoul(argv[3], NULL, 16);

#ifdef CONFIG_HAS_DATAFLASH
	if (addr_dataflash(addr1) | addr_dataflash(addr2)){
		puts ("Comparison with DataFlash space not supported.\n\r");
		return 0;
	}
#endif

	ngood = 0;

	while (count-- > 0) {
		if (size == 4) {
			ulong word1 = *(ulong *)addr1;
			ulong word2 = *(ulong *)addr2;
			if (word1 != word2) {
				printf("word at 0x%08lx (0x%08lx) "
					"!= word at 0x%08lx (0x%08lx)\n",
					addr1, word1, addr2, word2);
				rcode = 1;
				break;
			}
		}
		else if (size == 2) {
			ushort hword1 = *(ushort *)addr1;
			ushort hword2 = *(ushort *)addr2;
			if (hword1 != hword2) {
				printf("halfword at 0x%08lx (0x%04x) "
					"!= halfword at 0x%08lx (0x%04x)\n",
					addr1, hword1, addr2, hword2);
				rcode = 1;
				break;
			}
		}
		else {
			u_char byte1 = *(u_char *)addr1;
			u_char byte2 = *(u_char *)addr2;
			if (byte1 != byte2) {
				printf("byte at 0x%08lx (0x%02x) "
					"!= byte at 0x%08lx (0x%02x)\n",
					addr1, byte1, addr2, byte2);
				rcode = 1;
				break;
			}
		}
		ngood++;
		addr1 += size;
		addr2 += size;
	}

	printf("Total of %ld %s%s were the same\n",
		ngood, size == 4 ? "word" : size == 2 ? "halfword" : "byte",
		ngood == 1 ? "" : "s");
	return rcode;
}

int do_mem_cp ( cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{
	ulong	addr, dest, count;
	int	size;

	if (argc != 4) {
		printf ("Usage:\n%s\n", cmdtp->usage);
		return 1;
	}

	/* Check for size specification.
	*/
	if ((size = cmd_get_data_size(argv[0], 4)) < 0)
		return 1;

	addr = simple_strtoul(argv[1], NULL, 16);
	addr += base_address;

	dest = simple_strtoul(argv[2], NULL, 16);
	dest += base_address;

	count = simple_strtoul(argv[3], NULL, 16);

	if (count == 0) {
		puts ("Zero length ???\n");
		return 1;
	}

#ifndef CFG_NO_FLASH
	/* check if we are copying to Flash */
	if ( (addr2info(dest) != NULL)
#ifdef CONFIG_HAS_DATAFLASH
	   && (!addr_dataflash(addr))
#endif
	   ) {
		int rc;

		puts ("Copy to Flash... ");

		rc = flash_write ((uchar *)addr, dest, count*size);
		if (rc != 0) {
			flash_perror (rc);
			return (1);
		}
		puts ("done\n");
		return 0;
	}
#endif

#if (CONFIG_COMMANDS & CFG_CMD_MMC)
	if (mmc2info(dest)) {
		int rc;

		puts ("Copy to MMC... ");
		switch (rc = mmc_write ((uchar *)addr, dest, count*size)) {
		case 0:
			putc ('\n');
			return 1;
		case -1:
			puts ("failed\n");
			return 1;
		default:
			printf ("%s[%d] FIXME: rc=%d\n",__FILE__,__LINE__,rc);
			return 1;
		}
		puts ("done\n");
		return 0;
	}

	if (mmc2info(addr)) {
		int rc;

		puts ("Copy from MMC... ");
		switch (rc = mmc_read (addr, (uchar *)dest, count*size)) {
		case 0:
			putc ('\n');
			return 1;
		case -1:
			puts ("failed\n");
			return 1;
		default:
			printf ("%s[%d] FIXME: rc=%d\n",__FILE__,__LINE__,rc);
			return 1;
		}
		puts ("done\n");
		return 0;
	}
#endif

#ifdef CONFIG_HAS_DATAFLASH
	/* Check if we are copying from RAM or Flash to DataFlash */
	if (addr_dataflash(dest) && !addr_dataflash(addr)){
		int rc;

		puts ("Copy to DataFlash... ");

		rc = write_dataflash (dest, addr, count*size);

		if (rc != 1) {
			dataflash_perror (rc);
			return (1);
		}
		puts ("done\n");
		return 0;
	}

	/* Check if we are copying from DataFlash to RAM */
	if (addr_dataflash(addr) && !addr_dataflash(dest) && (addr2info(dest)==NULL) ){
		int rc;
		rc = read_dataflash(addr, count * size, (char *) dest);
		if (rc != 1) {
			dataflash_perror (rc);
			return (1);
		}
		return 0;
	}

	if (addr_dataflash(addr) && addr_dataflash(dest)){
		puts ("Unsupported combination of source/destination.\n\r");
		return 1;
	}
#endif

	while (count-- > 0) {
		if (size == 4)
			*((ulong  *)dest) = *((ulong  *)addr);
		else if (size == 2)
			*((ushort *)dest) = *((ushort *)addr);
		else
			*((u_char *)dest) = *((u_char *)addr);
		addr += size;
		dest += size;
	}
	return 0;
}

int do_mem_base (cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{
	if (argc > 1) {
		/* Set new base address.
		*/
		base_address64 = simple_strtoull(argv[1], NULL, 16);
        base_address = (uint32_t)base_address64;
	}
	/* Print the current base address.
	*/
	printf("Base Address64: 0x%16qx, ", base_address64);
	printf("Base Address: 0x%08lx\n", base_address);
	return 0;
}

int do_mem_loop (cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{
	ulong	addr, length, i, junk;
	int	size;
	volatile uint	*longp;
	volatile ushort *shortp;
	volatile u_char	*cp;

	if (argc < 3) {
		printf ("Usage:\n%s\n", cmdtp->usage);
		return 1;
	}

	/* Check for a size spefication.
	 * Defaults to long if no or incorrect specification.
	 */
	if ((size = cmd_get_data_size(argv[0], 4)) < 0)
		return 1;

	/* Address is always specified.
	*/
	addr = simple_strtoul(argv[1], NULL, 16);

	/* Length is the number of objects, not number of bytes.
	*/
	length = simple_strtoul(argv[2], NULL, 16);

	/* We want to optimize the loops to run as fast as possible.
	 * If we have only one object, just run infinite loops.
	 */
	if (length == 1) {
		if (size == 4) {
			longp = (uint *)addr;
			for (;;)
				i = *longp;
		}
		if (size == 2) {
			shortp = (ushort *)addr;
			for (;;)
				i = *shortp;
		}
		cp = (u_char *)addr;
		for (;;)
			i = *cp;
	}

	if (size == 4) {
		for (;;) {
			longp = (uint *)addr;
			i = length;
			while (i-- > 0)
				junk = *longp++;
		}
	}
	if (size == 2) {
		for (;;) {
			shortp = (ushort *)addr;
			i = length;
			while (i-- > 0)
				junk = *shortp++;
		}
	}
	for (;;) {
		cp = (u_char *)addr;
		i = length;
		while (i-- > 0)
			junk = *cp++;
	}
}


/**********************/
typedef struct {
  char        *name;
  uint64_t    base;
  uint64_t    extent;
} ddrMemoryRange;

#define Kb (1024UL)
#define Mb (1024UL*1024UL)
#define Gb (1024UL*Mb)

ddrMemoryRange memoryMap[4] =
                 {{"Lower 256Mb",  0x0000000000000000LL, 256*Mb},
                  {"Middle 256Mb", 0x0000000410000000LL, 256*Mb},
                  {"Upper 15.5Gb", 0x0000000020000000LL, 512*Mb},
                  {NULL, 0, 0}};

/*
 * Portable memory test code
 * Original paper: Michael Barr
 * Contributors: rogerliu
 * Date: 2005-05-09
 */
typedef uint64_t datum;    /* Set the data bus width to 64 bits. */


/**********************************************************************
 *
 * Function:    memTestDataBus()
 *
 * Description: Test the data bus wiring in a memory region by
 *              performing a walking 1's test at a fixed address
 *              within that region.  The address (and hence the
 *              memory region) is selected by the caller.
 *
 * Notes:       
 *
 * Returns:     0 if the test succeeds.  
 *              A non-zero result is the first pattern that failed.
 *
 **********************************************************************/
int memTestDataBus(volatile datum address)
{
    uint32_t pattern;

    /*
     * Perform a walking 1's test at the given address.
     */
    for (pattern = 1; pattern != 0; pattern <<= 1)
    {
        /*
         * Write the test pattern.
         */
        octeon_write64(address, pattern);

        /*
         * Read it back (immediately is okay for this test).
         */
        if (octeon_read64(address) != pattern) 
            return 1;
    }

    return 0;

}   /* memTestDataBus() */

/**********************************************************************
 *
 * Function:    memTestAddressBus()
 *
 * Description: Test the address bus wiring in a memory region by
 *              performing a walking 1's test on the relevant bits
 *              of the address and checking for aliasing. This test
 *              will find single-bit address failures such as stuck
 *              -high, stuck-low, and shorted pins.  The base address
 *              and size of the region are selected by the caller.
 *
 * Notes:       For best results, the selected base address should
 *              have enough LSB 0's to guarantee single address bit
 *              changes.  For example, to test a 64-Kbyte region, 
 *              select a base address on a 64-Kbyte boundary.  Also, 
 *              select the region size as a power-of-two--if at all 
 *              possible.
 *
 * Returns:     NULL if the test succeeds.  
 *              A non-zero result is the first address at which an
 *              aliasing problem was uncovered.  By examining the
 *              contents of memory, it may be possible to gather
 *              additional information about the problem.
 *
 **********************************************************************/
int memTestAddressBus(volatile datum baseAddress, volatile datum nBytes)
{
    uint64_t addressMax = baseAddress + nBytes;
    uint64_t offset;
    uint64_t testOffset;

    datum pattern     = (datum) 0xAAAAAAAAAAAAAAAA;
    datum antipattern = (datum) 0x5555555555555555;

    /*
     * Write the default pattern at each of the power-of-two offsets.
     */
    for (offset = sizeof(datum); (baseAddress+offset) < addressMax; offset <<= 1)
        octeon_write64(baseAddress+offset, pattern);

    /* 
     * Check for address bits stuck high.
     */
    testOffset = 0;
    octeon_write64(baseAddress, antipattern);

    for (offset = sizeof(datum); (baseAddress+offset) < addressMax; offset <<= 1)
    {
        if (octeon_read64((baseAddress+offset)) != pattern)
            return 1;
    }

    octeon_write64((baseAddress+testOffset), pattern);

    /*
     * Check for address bits stuck low or shorted.
     */
    for (testOffset  = sizeof(datum); (baseAddress+testOffset ) < addressMax; testOffset  <<= 1)
    {
        octeon_write64(baseAddress+testOffset, antipattern);

		if (octeon_read64(baseAddress) != pattern)
			return 1;

	    for (offset = sizeof(datum); (baseAddress+offset) < addressMax; offset <<= 1)
	        {
	            if ((octeon_read64(baseAddress+offset) != pattern) && (offset != testOffset))
	                return 1;
	        }

        octeon_write64(baseAddress+testOffset, pattern);
    }

    return 0;

}   /* memTestAddressBus() */

/**********************************************************************
 *
 * Function:    memTestDevice()
 *
 * Description: Test the integrity of a physical memory device by
 *              performing an increment/decrement test over the
 *              entire region.  In the process every storage bit 
 *              in the device is tested as a zero and a one.  The
 *              base address and the size of the region are
 *              selected by the caller.
 *
 * Notes:       
 *
 * Returns:     NULL if the test succeeds.
 *
 *              A non-zero result is the first address at which an
 *              incorrect value was read back.  By examining the
 *              contents of memory, it may be possible to gather
 *              additional information about the problem.
 *
 **********************************************************************/
int memTestDevice(volatile datum baseAddress, datum nBytes)	
{
    uint64_t offset;
    uint64_t nWords = nBytes / sizeof(datum);
    datum pattern;
    datum antipattern;


    /*
     * Fill memory with a known pattern.
     */
    for (pattern = 1, offset = 0; offset < nWords; pattern++, offset+=sizeof(datum))
        octeon_write64(baseAddress+offset, pattern);

    /*
     * Check each location and invert it for the second pass.
     */
    for (pattern = 1, offset = 0; offset < nWords; pattern++, offset+=sizeof(datum))
    {
        if (octeon_read64(baseAddress+offset) != pattern)
        {
        	printf("error : wrong pattern in ");
    		printf("baseAddress[offset]= %qx , offset %qx, octeon_read64(baseAddress)=%qx, pattern=%qx\n", baseAddress+offset, offset, octeon_read64(baseAddress+offset), pattern);
            return 1;
        }

        antipattern = ~pattern;
        octeon_write64(baseAddress+offset, antipattern);
    }

    /*
     * Check each location for the inverted pattern and zero it.
     */
    for (pattern = 1, offset = 0; offset < nWords; pattern++, offset+=sizeof(datum))
    {
        antipattern = ~pattern;
        if (octeon_read64(baseAddress+offset) != antipattern)
        {
        	printf("error : wrong pattern in ");
    		printf("baseAddress[offset]= %qx , offset %qx, octeon_read64(baseAddress)=%qx, antipattern=%qx\n", baseAddress+offset, offset, octeon_read64(baseAddress+offset), antipattern);
            return 1;
        }
    }

    return 0;

}   /* memTestDevice() */


/*
 * Perform a memory test. A more complete alternative test can be
 * configured using CFG_ALT_MEMTEST. The complete test loops until
 * interrupted by ctrl-c or by a failure of one of the sub-tests.
 */

#define CFG_ALT_MEMTEST
int do_mem_mtest (cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{
#if 0
	uint64_t	 start, end;

	if (argc > 2) {
		end = simple_strtoul(argv[2], NULL, 16);
	} else {
		end = memoryMap[1].base + memoryMap[1].extent;
	}
	
	if (argc > 1) {
		start = simple_strtoul(argv[1], NULL, 16);
	} else {
		start = memoryMap[0].base + 0x100000;
	}
#endif


	uint64_t	 m_start;
	
	printf ("Testing memory ...\n\n");
	
		m_start = memoryMap[0].base + 0x100000;
		printf ("Testing first 128M memory :\n");
		printf ("Testing from %qx to %qx.\n",m_start,m_start+(memoryMap[0].extent) /2 - 0x100000 -1);
		if (memTestDataBus(m_start) != 0) {
			printf("Data bus test error!\n");
			return 1;
		}
		printf("Data bus test passed.\n");
		if ((memTestAddressBus(m_start,  (memoryMap[0].extent) /2 - 0x100000) != 0) ) {
			printf("Address bus test error!\n");
			return 2;
		}
		printf("Address bus test passed.\n");
		if((memTestDevice(m_start, (memoryMap[0].extent) /2 - 0x100000) != 0) ) {
			printf("Device test error!\n");
			return 3;
		}
		printf("Device test passed.\n\n");


	printf("All memory test passed.\n\n");
	return 0;

}


int do_flash_rw_test(void) {
	int	i, result;
	unsigned char	*flash_addr, *ram_addr;
	unsigned char	readback;
	char	cmd[100];

	printf("[flash r/w test]\n");

	flash_addr = (ulong *)FLASHTEST_START;
	ram_addr = (ulong *)CFG_LOAD_ADDR;

	memcpy(ram_addr, flash_addr, FLASHTEST_END-FLASHTEST_START+1);
	if(memcmp(ram_addr, flash_addr, FLASHTEST_END-FLASHTEST_START+1)!=0)
	{
		printf("flash read fail !\n");
		return 1;
	}
	printf("flash read pass.\n");
		
	memset(cmd, 0x0, 100);
	sprintf(cmd, "protect off 0x%x 0x%x", FLASHTEST_START, FLASHTEST_END);
	printf("%s\n", cmd);
	run_command(cmd, 0);
	memset(cmd, 0x0, 100);
	sprintf(cmd, "erase 0x%x 0x%x", FLASHTEST_START, FLASHTEST_END);
	printf("%s\n", cmd);
	run_command(cmd, 0);
	
	result = 1;
	for (i=0;i<(FLASHTEST_END-FLASHTEST_START+1);i++){
		readback = *flash_addr;
		flash_addr++;
		if(readback != 0xFF) {
			result = 0;
			break;
		}
	}

	flash_addr = (ulong *)FLASHTEST_START;
	memset(cmd, 0x0, 100);
	sprintf(cmd, "cp.b 0x100000 0x%x 0x%x", FLASHTEST_START, FLASHTEST_END-FLASHTEST_START+1);
	printf("%s\n", cmd);
	run_command(cmd, 0);

	if(memcmp(ram_addr, flash_addr, FLASHTEST_END-FLASHTEST_START+1)!=0)
	{		
		printf("flash write fail !\n");
		return 1;
	}
	printf("flash write pass.\n");

	if ( 0 == result )
	{
		printf("flash erase fail !\n");
		return 1;
	}
	printf("flash erase pass.\n");

	printf("flash r/w test pass !\n");
	return 0;
	
}



/* Modify memory.
 *
 * Syntax:
 *	mm{.b, .w, .l} {addr}
 *	nm{.b, .w, .l} {addr}
 */
static int
mod_mem(cmd_tbl_t *cmdtp, int incrflag, int flag, int argc, char *argv[])
{
	ulong	addr, i;
	int	nbytes, size;
	extern char console_buffer[];

	if (argc != 2) {
		printf ("Usage:\n%s\n", cmdtp->usage);
		return 1;
	}

#ifdef CONFIG_BOOT_RETRY_TIME
	reset_cmd_timeout();	/* got a good command to get here */
#endif
	/* We use the last specified parameters, unless new ones are
	 * entered.
	 */
	addr = mm_last_addr;
	size = mm_last_size;

	if ((flag & CMD_FLAG_REPEAT) == 0) {
		/* New command specified.  Check for a size specification.
		 * Defaults to long if no or incorrect specification.
		 */
		if ((size = cmd_get_data_size(argv[0], 4)) < 0)
			return 1;

		/* Address is specified since argc > 1
		*/
		addr = simple_strtoul(argv[1], NULL, 16);
		addr += base_address;
	}

#ifdef CONFIG_HAS_DATAFLASH
	if (addr_dataflash(addr)){
		puts ("Can't modify DataFlash in place. Use cp instead.\n\r");
		return 0;
	}
#endif

	/* Print the address, followed by value.  Then accept input for
	 * the next value.  A non-converted value exits.
	 */
	do {
		printf("%08lx:", addr);
		if (size == 4)
			printf(" %08x", *((uint   *)addr));
		else if (size == 2)
			printf(" %04x", *((ushort *)addr));
		else
			printf(" %02x", *((u_char *)addr));

		nbytes = readline (" ? ");
		if (nbytes == 0 || (nbytes == 1 && console_buffer[0] == '-')) {
			/* <CR> pressed as only input, don't modify current
			 * location and move to next. "-" pressed will go back.
			 */
			if (incrflag)
				addr += nbytes ? -size : size;
			nbytes = 1;
#ifdef CONFIG_BOOT_RETRY_TIME
			reset_cmd_timeout(); /* good enough to not time out */
#endif
		}
#ifdef CONFIG_BOOT_RETRY_TIME
		else if (nbytes == -2) {
			break;	/* timed out, exit the command	*/
		}
#endif
		else {
			char *endp;
			i = simple_strtoul(console_buffer, &endp, 16);
			nbytes = endp - console_buffer;
			if (nbytes) {
#ifdef CONFIG_BOOT_RETRY_TIME
				/* good enough to not time out
				 */
				reset_cmd_timeout();
#endif
				if (size == 4)
					*((uint   *)addr) = i;
				else if (size == 2)
					*((ushort *)addr) = i;
				else
					*((u_char *)addr) = i;
				if (incrflag)
					addr += size;
			}
		}
	} while (nbytes);

	mm_last_addr = addr;
	mm_last_size = size;
	return 0;
}

#ifndef CONFIG_CRC32_VERIFY

int do_mem_crc (cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{
	ulong addr, length;
	ulong crc;
	ulong *ptr;

	if (argc < 3) {
		printf ("Usage:\n%s\n", cmdtp->usage);
		return 1;
	}

	addr = simple_strtoul (argv[1], NULL, 16);
	addr += base_address;

	length = simple_strtoul (argv[2], NULL, 16);

	crc = crc32 (0, (const uchar *) addr, length);

	printf ("CRC32 for %08lx ... %08lx ==> %08lx\n",
			addr, addr + length - 1, crc);

	if (argc > 3) {
		ptr = (ulong *) simple_strtoul (argv[3], NULL, 16);
		*ptr = crc;
	}

	return 0;
}

#else	/* CONFIG_CRC32_VERIFY */

int do_mem_crc (cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{
	ulong addr, length;
	ulong crc;
	ulong *ptr;
	ulong vcrc;
	int verify;
	int ac;
	char **av;

	if (argc < 3) {
  usage:
		printf ("Usage:\n%s\n", cmdtp->usage);
		return 1;
	}

	av = argv + 1;
	ac = argc - 1;
	if (strcmp(*av, "-v") == 0) {
		verify = 1;
		av++;
		ac--;
		if (ac < 3)
			goto usage;
	} else
		verify = 0;

	addr = simple_strtoul(*av++, NULL, 16);
	addr += base_address;
	length = simple_strtoul(*av++, NULL, 16);

	crc = crc32(0, (const uchar *) addr, length);

	if (!verify) {
		printf ("CRC32 for %08lx ... %08lx ==> %08lx\n",
				addr, addr + length - 1, crc);
		if (ac > 2) {
			ptr = (ulong *) simple_strtoul (*av++, NULL, 16);
			*ptr = crc;
		}
	} else {
		vcrc = simple_strtoul(*av++, NULL, 16);
		if (vcrc != crc) {
			printf ("CRC32 for %08lx ... %08lx ==> %08lx != %08lx ** ERROR **\n",
					addr, addr + length - 1, crc, vcrc);
			return 1;
		}
	}

	return 0;

}
#endif	/* CONFIG_CRC32_VERIFY */

#include "octeon_boot.h"
#include "cvmx-bootmem-shared.h"
int do_bootmem_named_alloc (cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{
    char *name;
    uint64_t addr = 0;
    uint64_t max_addr = 0;
    uint64_t size;
    char buf[20];

    setenv("named_block_addr", NULL);
    setenv("named_block_size", NULL);

    if (argc < 3)
    {
		printf ("Usage:\n%s\n", cmdtp->usage);
		return 1;
    }
    name = argv[1];
    size = simple_strtoull(argv[2], NULL, 16);
    if (argc == 4)
    {
        addr = simple_strtoull(argv[3], NULL, 16);
        max_addr = addr + size;
    }

#ifdef DEBUG
    printf("Name: %s, size: 0x%qx, addr: 0x%qx\n", name, size, addr);
#endif

    if (!glob_phy_desc_ptr)
    {
        printf("Error, bootmem alloc not initialized!\n");
        return(1);
    }
    addr = octeon_phy_mem_named_block_alloc(glob_phy_desc_ptr, size, addr, max_addr, 0, name);
    if (!addr)
    {
        printf("Named allocation failed!\n");
        return(1);
    }
    printf("Allocated 0x%qx bytes at address: 0x%qx, name: %s\n", size, addr, name);

    sprintf(buf, "0x%qX", addr);
    setenv("named_block_addr", buf);
    sprintf(buf, "0x%qX", size);
    setenv("named_block_size", buf);


#ifdef DEBUG
    printf("\n\n===========================================\n");
    printf("Dump of named blocks:\n");
    octeon_phy_mem_named_block_print(glob_phy_desc_ptr);
    octeon_phy_mem_list_print(glob_phy_desc_ptr);
#endif

    return(0);
}

int do_bootmem_named_free (cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{
    if (argc < 2)
    {
		printf ("Usage:\n%s\n", cmdtp->usage);
		return 1;
    }
    if (!octeon_phy_mem_named_block_free(glob_phy_desc_ptr, argv[1]))
    {
        printf("Error freeing block: %s\n", argv[1]);
        return(1);
    }
    return(0);
}

int do_bootmem_named_print (cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{
    octeon_phy_mem_named_block_print(glob_phy_desc_ptr);
    return(0);
}

/**************************************************/
#if (CONFIG_COMMANDS & CFG_CMD_MEMORY)
U_BOOT_CMD(
	md,     3,     1,      do_mem_md,
	"md      - memory display\n",
	"[.b, .w, .l] address [# of objects]\n    - memory display\n"
);


U_BOOT_CMD(
	mm,     2,      1,       do_mem_mm,
	"mm      - memory modify (auto-incrementing)\n",
	"[.b, .w, .l] address\n" "    - memory modify, auto increment address\n"
);


U_BOOT_CMD(
	nm,     2,	    1,     	do_mem_nm,
	"nm      - memory modify (constant address)\n",
	"[.b, .w, .l] address\n    - memory modify, read and keep address\n"
);

U_BOOT_CMD(
	mw,    4,    1,     do_mem_mw,
	"mw      - memory write (fill)\n",
	"[.b, .w, .l] address value [count]\n    - write memory\n"
);

U_BOOT_CMD(
	cp,    4,    1,    do_mem_cp,
	"cp      - memory copy\n",
	"[.b, .w, .l] source target count\n    - copy memory\n"
);

U_BOOT_CMD(
	cmp,    4,     1,     do_mem_cmp,
	"cmp     - memory compare\n",
	"[.b, .w, .l] addr1 addr2 count\n    - compare memory\n"
);

#ifndef CONFIG_CRC32_VERIFY

U_BOOT_CMD(
	crc32,    4,    1,     do_mem_crc,
	"crc32   - checksum calculation\n",
	"address count [addr]\n    - compute CRC32 checksum [save at addr]\n"
);

#else	/* CONFIG_CRC32_VERIFY */

U_BOOT_CMD(
	crc32,    5,    1,     do_mem_crc,
	"crc32   - checksum calculation\n",
	"address count [addr]\n    - compute CRC32 checksum [save at addr]\n"
	"-v address count crc\n    - verify crc of memory area\n"
);

#endif	/* CONFIG_CRC32_VERIFY */

U_BOOT_CMD(
	base,    2,    1,     do_mem_base,
	"base    - print or set address offset\n",
	"\n    - print address offset for memory commands\n"
	"base off\n    - set address offset for memory commands to 'off'\n"
);
U_BOOT_CMD(
	read64,    2,    1,     do_read64,
	"read64    - read 64 bit word from 64 bit address\n",
	"\n    - read 64 bit word from 64 bit address\n"
	"read64 addr\n    - read 64 bit word from 64 bit address\n"
);
U_BOOT_CMD(
	write64,    4,    1,     do_write64,
	"write64    - write 64 bit word to 64 bit address\n",
	"\n    - write 64 bit word to 64 bit address\n"
	"write64 addr val\n    - write 64 bit word to 64 bit address\n"
);
U_BOOT_CMD(
	namedalloc,    4,    0,     do_bootmem_named_alloc,
	"namedalloc    - Allocate a named bootmem block\n",
	"\n    - Allocate a named bootmem block\n"
	"namedalloc name size [address]\n"
    "    - Allocate a named bootmem block with a given name and size at an\n"
    "      optional fixed address.  Sets environment variables named_block_addr,\n"
    "      named_block_size to address and size of block allocated.\n"
);

U_BOOT_CMD(
	namedfree,    2,    0,     do_bootmem_named_free,
	"namedfree    - Free a named bootmem block\n",
	"\n    - Free a named bootmem block\n"
	"namedfree name\n"
    "    - Free a named bootmem block.\n"
);
U_BOOT_CMD(
	namedprint,    1,    0,     do_bootmem_named_print,
	"namedprint    - Print list of named bootmem blocks\n",
	"\n    - Print list of named bootmem blocks\n"
	"namedprint\n"
    "    - Print list of named bootmem blocks.\n"
);


U_BOOT_CMD(
	read64b,    2,    1,     do_read64b,
	"read64b    - read 8 bit word from 64 bit address\n",
	"\n    - read 8 bit word from 64 bit address \n"
	"read64b addr\n    - read 8 bit word from 64 bit address\n"
);
U_BOOT_CMD(
	read64l,    2,    1,     do_read64l,
	"read64l    - read 32 bit word from 64 bit address\n",
	"\n    - read 32 bit word from 64 bit address \n"
	"read64l addr\n    - read 32 bit word from 64 bit address\n"
);
U_BOOT_CMD(
	write64b,    4,    1,     do_write64b,
	"write64b    - write 8 bit word to 64 bit address\n",
	"\n    - write 8 bit word to 64 bit address\n"
	"write64b addr val\n    - write 8 bit word to 64 bit address\n"
);
U_BOOT_CMD(
	write64l,    4,    1,     do_write64l,
	"write64l    - write 32 bit word to 64 bit address\n",
	"\n    - write 32 bit word to 64 bit address\n"
	"write64l addr val\n    - write 32 bit word to 64 bit address\n"
);

U_BOOT_CMD(
	read_cmp,    4,    1,     do_read_cmp,
	"read_cmp    - read and compare memory to val\n",
	"\n    - read and compare memory to val\n"
	"read_cmp addr val\n    - read and compare memory to val\n"
);

U_BOOT_CMD(
	loop,    3,    1,    do_mem_loop,
	"loop    - infinite loop on address range\n",
	"[.b, .w, .l] address number_of_objects\n"
	"    - loop on a set of addresses\n"
);

U_BOOT_CMD(
	mtest,    2,    1,     do_mem_mtest,
	"mtest   - simple RAM test\n",
	"\n"
	"\t simple RAM read/write test\n"
);

U_BOOT_CMD(
	ftest,    1,    1,     do_flash_rw_test,
	"",
	"\n"
	"\t simple flash read/write test\n"
);

#endif
#endif	/* CFG_CMD_MEMORY */
